/*
  Userland OOM-averting program.
  Copyright (C) 2010 Deekoo L.

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */

#include <sys/types.h>
#include <sys/stat.h>
#include <sys/select.h>
#include <fcntl.h>
#include <stdio.h>
#include <dirent.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>

/* FAIL_MEM is the amount of memory to return as free in the event it can't
  figure out how much memory is actually free.  By default, this is zero,
  as if /proc/meminfo can't be opened you may well have an oom.  However,
  if /proc/meminfo can't be opened oomer (at least in userspace) will probably
  also fail to open pid cmdlines. */

#define FAIL_MEM 0


/*TODO:
        configfs
        figure out where to hook oomer for best results.
        figure out how to check free/low memory.
 */
        
DIR *proc_dir;
char datablock[PATH_MAX];
int recentkillptr = 0;
time_t now=0;
int numrecentkills = 0;
#define MAXRECENTKILLS 32
struct {
        pid_t pid;
        int signal;
        time_t timekilled;
        } recentkills[MAXRECENTKILLS];

typedef enum {
        oomer_pid,
        oomer_process_filename,
        } oomer_pat_t;

typedef struct {
        int len;
        char *str;
} oomer_str;


void clean_recent_list(void)
{
  int k = numrecentkills;
  time_t thresh = now-30;
  int orecentkills = numrecentkills;
  while (k) {
    k--;
    if ((recentkills[k].timekilled<thresh) || (kill(recentkills[k].pid,0)<0)){
      numrecentkills--;
      recentkills[k] = recentkills[numrecentkills];
    }
  }
  if (orecentkills>numrecentkills) {
    recentkillptr = numrecentkills;
  }
}

struct oomer_script_line {
        int signal;
        oomer_pat_t type;
        union {
                oomer_str s;
                int pid;
        } pat;
};

struct oomer_script_block {
    size_t freemem;
    int numscripts;
    struct oomer_script_line *scripts;
};

int numblocks = 0;
struct oomer_script_block *blocks = NULL; // should be sorted when modified.

struct oomer_script_block *oomer_block(size_t freemem)
{
  int x = 0;
  while (x<numblocks) {
    if (blocks[x].freemem==freemem) return &blocks[x];
    x++;
  }
  return NULL;
}

struct oomer_script_block *oomer_add_block(size_t freemem)
{
  int x = 0;
  int pos = 0;
  while (x<numblocks) {
    if (blocks[x].freemem==freemem) {
      return &blocks[x];
    } else if (blocks[x].freemem>freemem) {
      x++;
      pos = x;
    } else {
      pos = x;
      x = numblocks;
    }
  }
  struct oomer_script_block *newblocks = realloc(blocks, sizeof(struct oomer_script_block)*(numblocks+1));
  if (newblocks) {
    blocks = newblocks;
    x = pos;
    while (x<numblocks) {
      blocks[x+1] = blocks[x];
      x++;
    }
    numblocks++;
    blocks[pos].freemem = freemem;
    blocks[pos].numscripts = 0;
    blocks[pos].scripts = NULL;
  }
  return &blocks[pos];
}

void oomer_add_script(struct oomer_script_block *block, int signal, oomer_pat_t type, void *pat)
{
  struct oomer_script_line *newscripts = realloc(block->scripts,
    sizeof(struct oomer_script_line)*(block->numscripts+1));
  if (newscripts) {
    block->scripts = newscripts;
    block->scripts[block->numscripts].type = type;
    block->scripts[block->numscripts].signal = signal;
    switch (type) {
      case oomer_process_filename:
        block->scripts[block->numscripts].pat.s = *((oomer_str *)pat);
        break;
      default:
        exit(-1);
    }
    block->numscripts++;
  }
}

int parse_line(char *line)
{
  warn("XXX parse_line unwritten.\n");
}

int valid_procname(const char *fn)
{
  int x = 0;
  int a = 0;
  while (fn[x]) {
    if (fn[x]<'0' || fn[x]>'9') return 0;
    a *= 10;
    a += fn[x]-'0';
    x++;
  }
  return a;
}

int recentlykilled(pid_t pid,int signal)
{
  int x = 0;
  while (x<numrecentkills) {
    if (recentkills[x].pid==pid && recentkills[x].signal==signal) return 1;
    x++;
  }
  return 0;
}

void kill_unless_protected(pid_t pid, int signal)
{
  if (pid==1) return;
  recentkills[recentkillptr].pid = pid;
  recentkills[recentkillptr].timekilled = now;
  recentkills[recentkillptr].signal = signal;
  recentkillptr++;
  numrecentkills++;
  if (recentkillptr>=MAXRECENTKILLS) {
    recentkillptr = 0;
    numrecentkills = MAXRECENTKILLS;
  }
  kill(pid, signal);
  int mesg[4];
  mesg[0] = *((int *)"KILL");
  mesg[1] = pid;
  mesg[2] = signal;
  mesg[3] = *((int *)"...\n");
  write(1,&mesg[0],16);
  //fprintf(stderr, "Killed %d,%d\n",pid,signal);
}

inline void kill_by_process(struct oomer_script_line *line, int pid)
{
  snprintf(&datablock[0],PATH_MAX,"/proc/%d/cmdline",pid);
  int cmdline = open(&datablock[0],O_RDONLY);
  if (cmdline>=0) {
    int cmdpos = read(cmdline,&datablock[0],PATH_MAX);
    if (cmdpos>=0) {
      int x = 0;
      while (x<sizeof(datablock)) {
        if (datablock[x]) {
          x++;
        } else {
          if (x-line->pat.s.len>=0) {
            if (!memcmp(line->pat.s.str, &datablock[x-line->pat.s.len],line->pat.s.len)) {
              kill_unless_protected(pid,line->signal);
            }
          //} else {
          //  fprintf(stderr,"DBG: str %s, len %d, x=%d\n",line->pat.s.str,line->pat.s.len,x);
          }
          close(cmdline);
          return;
        }
      }
    } // XXX
    close(cmdline);
  }
}

inline void kill_process_filename(struct oomer_script_line *line)
{
  rewinddir(proc_dir);
  struct dirent *pf = readdir(proc_dir);
  while (pf) {
    int pid;
    if (pid=valid_procname(&pf->d_name[0])) {
      if (!recentlykilled(pid,line->signal)) {
        kill_by_process(line,pid);
      }
    }
    pf = readdir(proc_dir);
  }
}

void run_script_line(struct oomer_script_line *line)
{
  switch (line->type) {
    case oomer_process_filename:
      kill_process_filename(line);
      break;
    default:
      warn("Unhandled oomer condition type.\n");
  }
}

size_t get_free_memory(void)
{
  int fd = open("/proc/meminfo",O_RDONLY);
  size_t ret = FAIL_MEM;
  if (fd>=0) {
    int nread = (read(fd,&datablock[0],PATH_MAX));
    if (nread>=0) {
      datablock[PATH_MAX-1] = '\0';
      if (nread<PATH_MAX-1) datablock[nread] = '\0';
      int memfree, buffers, cached;
      int res = sscanf(&datablock[0],"MemTotal: %d kB MemFree: %d kB Buffers: %d kB Cached: %d kB", &ret, &memfree, &buffers, &cached);
      if (res==4) {
        ret = buffers;
        ret += cached;
        ret += memfree;
        ret *= 1024;
      }
    }
    close(fd);
  }
  return ret;
}

void do_oomer()
{
  int x = 0;
  size_t current_free_memory = get_free_memory();
  while (x<numblocks) {
    if (blocks[x].freemem<current_free_memory) return;
    int y = 0;
    while (y<blocks[x].numscripts) {
      run_script_line(&blocks[x].scripts[y]);
      y++;
    }
    x++;
  }
}

void init()
{
  fcntl(1,F_SETFL,O_NONBLOCK);
  proc_dir = opendir("/proc");
  if (!proc_dir) {
    warn("Failed to open /proc.\n");
  }
}

void kill_by_name(int megs, int signal, char *prog)
{
  struct oomer_script_block *block = oomer_add_block(megs*1024*1024);
  oomer_str pat = { strlen(prog), prog };
  oomer_add_script(block,signal,oomer_process_filename,&pat);
}

int main(int *argc, char **argv)
{
  init();
  /* Change these as desired. */
  /* We kill gtk-gnash if available memory is less than 512 megs, and we
        kill it with SIGKILL because it's never doing anything important. */
  kill_by_name(512,SIGKILL,"gtk-gnash");
  /* Firefox, Konqueror, and the Gimp get -STOP'd at 400, 384, and 256 megs
        respectively.  We STOP instead of killing so you can resume them if
        there's any work you want to save in them. */
  kill_by_name(400,SIGSTOP,"firefox-bin");
  kill_by_name(384,SIGSTOP,"konqueror");
  kill_by_name(256,SIGSTOP,"gimp");
  struct timeval onesecond;
  while (1) {
    onesecond.tv_usec = onesecond.tv_sec = 1;
    select(0,NULL, NULL, NULL, &onesecond);
    now = time(NULL);
    clean_recent_list();
    do_oomer();
  }
}
